package net.sergeych.kiloparsec

import assertThrows
import io.ktor.server.engine.*
import io.ktor.server.netty.*
import kotlinx.coroutines.delay
import kotlinx.coroutines.launch
import kotlinx.coroutines.test.runTest
import net.sergeych.crypto2.initCrypto
import net.sergeych.kiloparsec.adapter.acceptTcpDevice
import net.sergeych.kiloparsec.adapter.connectTcpDevice
import net.sergeych.kiloparsec.adapter.setupWebsocketServer
import net.sergeych.kiloparsec.adapter.websocketClient
import net.sergeych.mp_logger.Log
import java.net.InetAddress
import kotlin.test.*

class ClientTest {

    class TestException : Exception("test1")

    @Test
    fun testAddresses() {
        println(InetAddress.getLocalHost())
        println(InetAddress.getByName("localhost"))
        println(InetAddress.getByName("127.0.0.1"))
        println(InetAddress.getByName("mail.ru"))
    }
    @Test
    fun testClient() = runTest {
        initCrypto()
        Log.connectConsole(Log.Level.DEBUG)
        data class Session(
            var data: String
        )

        val cmdSave by command<String,Unit>()
        val cmdLoad by command<Unit,String>()
        val cmdDrop by command<Unit,Unit>()
        val cmdException by command<Unit,Unit>()

        val cli = KiloInterface<Session>().apply {
            registerError { TestException() }
            onConnected { session.data = "start" }
            on(cmdSave) { session.data = it }
            on(cmdLoad) {
                session.data
            }
            on(cmdException) {
                throw TestException()
            }
            on(cmdDrop) {
                throw LocalInterface.BreakConnectionException()
            }
        }
        val server = KiloServer(cli, acceptTcpDevice(27101)) {
            Session("unknown")
        }

        val client = KiloClient<Unit>() {
            addErrors(cli)
            connect { connectTcpDevice("localhost:27101") }
        }
        delay(500)
        println(client.call(cmdLoad))

        assertEquals("start", client.call(cmdLoad))
        client.call(cmdSave, "foobar")
        assertEquals("foobar", client.call(cmdLoad))
//
        val res = kotlin.runCatching {  client.call(cmdException) }
        println(res.exceptionOrNull())
        assertIs<TestException>(res.exceptionOrNull())
        assertEquals("foobar", client.call(cmdLoad))

        assertThrows<RemoteInterface.ClosedException> { client.call(cmdDrop) }

        // reconnect?
        assertEquals("start", client.call(cmdLoad))

        server.close()
    }

    @Test
    fun webSocketTest() = runTest {
        initCrypto()
//        fun Application.
        val cmdClose by command<Unit,Unit>()
        val cmdGetFoo by command<Unit,String>()
        val cmdSetFoo by command<String,Unit>()
        val cmdCheckConnected by command<Unit,Boolean>()

        Log.connectConsole(Log.Level.DEBUG)

        data class Session(var foo: String="not set")
        var closeCounter = 0
        val serverInterface = KiloInterface<Session>().apply {
            var connectedCalled = false
            onConnected { connectedCalled = true }
            on(cmdGetFoo) { session.foo }
            on(cmdSetFoo) { session.foo = it }
            on(cmdCheckConnected) { connectedCalled }
            on(cmdClose) {
                throw LocalInterface.BreakConnectionException()
            }
        }

        val ns: NettyApplicationEngine = embeddedServer(Netty, port = 8080, host = "0.0.0.0", module = {
            setupWebsocketServer(serverInterface) { Session() }
        }).start(wait = false)

        val client = websocketClient<Unit>("ws://localhost:8080/kp")
        val states = mutableListOf<Boolean>()
        val collector = launch {
            client.connectedStateFlow.collect {
                println("got: $closeCounter/$it")
                states += it
                if( !it) { closeCounter++ }
            }
        }
        assertEquals(true, client.call(cmdCheckConnected))
        assertTrue { client.connectedStateFlow.value }
        assertEquals("not set", client.call(cmdGetFoo))
        client.call(cmdSetFoo, "foo")
        assertEquals("foo", client.call(cmdGetFoo))

        assertTrue { client.connectedStateFlow.value }
        assertThrows<RemoteInterface.ClosedException> {
            client.call(cmdClose)
        }

        // connection should now be closed
        assertFalse { client.connectedStateFlow.value }

        // this should be run on automatically reopen connection
        client.call(cmdSetFoo,"superbar")
        assertTrue { client.connectedStateFlow.value }
        assertEquals("superbar", client.call(cmdGetFoo))
        client.close()
        ns.stop()
        collector.cancel()
//        println("----= states: $states")
//        println("stopped server")
//        println("closed client")
    }
}